/***************************************************************************
                     args.cpp  -  arguments of the command line
                             -------------------
    begin                : Mon Aug 14 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#ifdef HAVE_GETOPT_H
  #include <getopt.h>
#endif

#include "args.h"
#include "global.h"
#include "common.h"

#include <string.h>
#include <stdlib.h>

static char optstring[]="z:oV:ecmdhf:s:p:bwg:vynSMa:iAI:k:E:rD";

#ifdef HAVE_GETOPT_H
static struct option const long_options[] =
{
  {"compress", required_argument, NULL, 'z'},
  {"overwrite", no_argument, NULL, 'o'},
  {"volume", required_argument, NULL, 'V'},
  {"waitvol", no_argument, NULL, 'w'},
  {"erase", no_argument, NULL, 'e'},
  {"encrypt", required_argument, NULL, 'E'},
  {"nocheck", no_argument, NULL, 'c'},
  {"allowmnt", no_argument, NULL, 'm'},
  {"nodesc", no_argument, NULL, 'd'},
  {"help", no_argument, NULL, 'h'},
  {"version", no_argument, NULL, 'v'},
  {"compilinfo", no_argument, NULL, 'i'},
  {"batch", no_argument, NULL, 'b'},
  {"finish", required_argument, NULL, 'f'},
  {"automnt", required_argument, NULL, 'a'},
  {"debug", required_argument, NULL, 'g'},
  {"nosync", no_argument, NULL, 'y'},
  {"nossl", no_argument, NULL, 'n'},
  {"nombr", no_argument, NULL, 'M'},
  {"simulate", no_argument, NULL, 'S'},
  {"saveall", no_argument, NULL, 'A'},
  {"server", required_argument, NULL, 's'},
  {"runserver", no_argument, NULL, 'r'},
  {"port", required_argument, NULL, 'p'},
  {"workadr", required_argument, NULL, 'I'},
  {"workport", required_argument, NULL, 'k'},
  {"nodetail", no_argument, NULL, 'D'},
  {NULL, 0, NULL, 0}
};
#endif

// =======================================================
void initOptions(COptions *options)
{
  memset(options, 0, sizeof(COptions));
  options->bUseSSL = OPTIONS_DEFAULT_SSL;
  options->bBackupMBR = OPTIONS_DEFAULT_BACKUP_MBR;
  options->bSimulateMode = OPTIONS_DEFAULT_SIMULATE_MODE;
  options->bOverwrite = OPTIONS_DEFAULT_OVERWRITE;
  options->bSaveFreeBlocks = OPTIONS_DEFAULT_SAVEFREEBLOCKS;
  options->qwSplitSize = OPTIONS_DEFAULT_SPLIT_SIZE * 1024;
  options->bEraseWithNull = OPTIONS_DEFAULT_ERASE_EMPTY;
#ifdef LEVELSUPPORT_GZIP
  options->dwCompression = OPTIONS_DEFAULT_COMPRESS;
#else
  options->dwCompression = COMPRESS_NONE;
#endif // LEVELSUPPORT_GZIP
  options->bCheckBeforeSaving = OPTIONS_DEFAULT_CHECK;
  options->bFailIfMounted = OPTIONS_DEFAULT_FAIL_IF_MOUNTED;
  options->bAskDesc = OPTIONS_DEFAULT_ASK_DESC;
  options->wFinish = OPTIONS_DEFAULT_FINISH;
  options->dwServerPort = OPTIONS_DEFAULT_SERVERPORT;
  options->bBatchMode = OPTIONS_DEFAULT_AUTOSTART;
  options->bSplitWait = OPTIONS_DEFAULT_SPLIT_WAIT;
  options->bSync = OPTIONS_DEFAULT_SYNC;
  options->dwDebugLevel = DEFAULT_DEBUG_LEVEL; // defined by configure.in
  options->bShowPartDetails = OPTIONS_DEFAULT_SHOW_PARTDETAILS;
  strncpy(options->szServerName, OPTIONS_DEFAULT_SERVERNAME, MAX_HOSTNAMESIZE);
  strncpy(options->szAutoMount, OPTIONS_DEFAULT_AUTOMOUNT, MAXPATHLEN);
  memset(options->cPassword, 0, MAX_PASSLEN);
  options->wEncryptAlgo = ENCRYPT_NONE;
}

// =======================================================
int processCommandLine(int argc, char **argv, COptions *options, BOOL *bStartServer, char *szServerName, WORD *wServerPort)
{
  int nOptCh;
  char szTemp[2048];

#ifdef HAVE_GETOPT_H
  while ((nOptCh = getopt_long (argc, argv, optstring, long_options, NULL)) != EOF)
#else
    while ((nOptCh = getopt (argc, argv, optstring)) != -1)
#endif
      {
	switch(nOptCh)
	  {
	  case 's': // partimaged server's ip addr
	    strncpy(options->szServerName, optarg, MAX_HOSTNAMESIZE);
	  case 'p': // partimaged server's listening port
	    options->dwServerPort = atol(optarg);
	    if ((options->dwServerPort < 1) || (options->dwServerPort > 65535))
	      {	
		fprintf(stderr, i18n("server's port must be between 1 and 65535\n"));
		return EXIT_FAILURE;
	      }
	    break;
	  case 'I': // partimage-work server's ip addr
	    strncpy(szServerName, optarg, MAX_HOSTNAMESIZE);
	    break;
	  case 'k': // partimage-work server's listening port
	    *wServerPort = atol(optarg);
	    if ((*wServerPort < 1)) // || (*wServerPort > 65535))
	      {	
		fprintf(stderr, i18n("server's port must be between 1 and "
				     "65535\n"));
		return EXIT_FAILURE;
	      }
	    break;
	  case 'E': // password for encryption
#if defined(LIBMCRYPT22) || defined(LIBMCRYPT24)
	    if (strlen(optarg) < MIN_PASSLEN)
	      {
		fprintf(stderr, i18n("The minimal password length is %lu\n"), (DWORD)MIN_PASSLEN);
		return EXIT_FAILURE;
	      }
	    strncpy((char*)options->cPassword, optarg, MAX_PASSLEN);
            options->wEncryptAlgo = ENCRYPT_AES;
#else
            fprintf(stderr, i18n("You cannot use emcryption since pi-newt"
               " wasn't compiled with crypting support\n"));
            return EXIT_FAILURE;
#endif
	    break;
	  case 'a': // auto mount options
	    strncpy(options->szAutoMount, optarg, MAXPATHLEN);
	    break;
	  case 'b': // batchmode: don't wait for user action (return)
	    options->bBatchMode = true;
	    break;
	  case 'w': // wait after each new colume
	    options->bSplitWait = true;
	    break;
	  case 'M': // do not create a backup of the MBR
	    options->bBackupMBR = false;
	    break;
	  case 'y': // no sync at the end
	    options->bSync = false;
	    break;
	  case 'z': // compression level
	    options->dwCompression = atol(optarg);
	    if ((options->dwCompression < 0) || (options->dwCompression > 3))
	      {	
		fprintf(stderr, i18n("Compression mode must be 0 (none), "
				     "1(lzo), 2(gzip), or 3(bzip2)\n"));
		return EXIT_FAILURE;
	      }
#ifndef LEVELSUPPORT_LZO
	    if (options->dwCompression == COMPRESS_LZO)
	      {
		fprintf(stderr, i18n("Cannot use %s compression level: it was disabled at compile time.\n"), "LZO");
		return EXIT_FAILURE;
	      }
#endif // LEVELSUPPORT_LZO
#ifndef LEVELSUPPORT_GZIP
	    if (options->dwCompression == COMPRESS_GZIP)
	      {
		fprintf(stderr, i18n("Cannot use %s compression level: it was disabled at compile time.\n"), "GZIP");
		return EXIT_FAILURE;
	      }
#endif // LEVELSUPPORT_GZIP
#ifndef LEVELSUPPORT_BZIP2
	    if (options->dwCompression == COMPRESS_BZIP2)
	      {
		fprintf(stderr, i18n("Cannot use %s compression level: it was disabled at compile time.\n"), "BZIP2");
		return EXIT_FAILURE;
	      }
#endif // LEVELSUPPORT_BZIP2
	    break;
	  
	  case 'f': // what to do when finished successfully ?
	    options->wFinish = atol(optarg);
	    if (options->wFinish > 3)
	      {
		fprintf(stderr, i18n("Finish mode must be 0 (wait), "
				     "1(quit), 2(reboot) or 3(halt)\n"));
		return EXIT_FAILURE;
	      }
	    break;
	  
	  case 'g': // change the debug level
	    options->dwDebugLevel = atol(optarg);
	    if ((options->dwDebugLevel < 0) || (options->dwDebugLevel > 10))
	      {
		fprintf(stderr, i18n("The debug level must be 0(none), "
				     "1(user), 2(developer) or 3(debugging)\n"));
		return EXIT_FAILURE;
	      }
	    break;
	  
	  case 'V': // split image into multiple files
	  
	    options->qwSplitSize = atol(optarg) * 1024;
	    fprintf (stderr, i18n("Volume size: %llu bytes (%ld KB)\n"),
		     options->qwSplitSize, atol(optarg));
	    break;
	  
	  case 'o': // overwrite existring image
	    options->bOverwrite = true;
	    break;
	  
	  case 'e': // erase empty blocks with zero bytes
	    options->bEraseWithNull = true;
	    break;
	  
	  case 'm': // don't fail if mounted
	    options->bFailIfMounted = false;
	    break;
	  
	  case 'c': // don't check partition before saving
	    options->bCheckBeforeSaving = false;
	    break;
	  
	  case 'd': // don't ask any description
	    options->bAskDesc = false;
	    break;
	  
	  case 'h': // help
	    formatUsage(szTemp, sizeof(szTemp));
	    printf("%s", szTemp);
	    return EXIT_SUCCESS;
	    break;
	  
	  case 'v': // version
	    printf(i18n("Partition Image version %s (distributed under the"
			" GNU GPL2).\n"), VERSION);
	    return EXIT_SUCCESS;
	    break;

	  case 'i': // compilation options
	    formatCompilOptions(szTemp, sizeof(szTemp));
	    printf("%s\n", szTemp);
	    return EXIT_SUCCESS;
	    break;

	  case 'n': // no ssl
	    options->bUseSSL = false;
	    break;

	  case 'S': // simulate
	    options->bSimulateMode = true;
	    break;

	  case 'A': // save all blocks
	    options->bSaveFreeBlocks = true;
	    break;

	  case 'r': // start the server
	    *bStartServer = true;
	    break;

	  case 'D': // show part details
	    options->bShowPartDetails = false;
	    break;
	  }
      }
  RETURN_int(0);
}

// =======================================================
void formatUsage(char *szText, int nMaxLen)
{
  int nLen;
  
  memset(szText, 0, nMaxLen);
  nLen = 0;
  nLen += snprintf(szText, nMaxLen-nLen, "===============================================================================\n");
  nLen += snprintf(szText, nMaxLen-nLen, 
		   i18n("Partition Image (http://www.partimage.org/) version %s [%s]\n"
			"---- distributed under the GPL 2 license (GNU General Public License) ----\n\n"
			"Supported file systems:....ext2fs, reiserfs, fat16, fat32, ntfs(exp), hpfs, jfs, xfs, ffs, hfs\n\n"
			"usage: partimage [options] <action> <device> <image_file>\n"
			"       partimage <imginfo/restmbr> <image_file>\n\n"
			"ex: partimage -z1 -o -d save /dev/hda12 /mnt/backup/redhat-6.2.partimg.gz\n"
			"ex: partimage restore /dev/hda13 /mnt/backup/suse-6.4.partimg\n"
			"ex: partimage restmbr /mnt/backup/debian-potato-2.2.partimg.bz2\n"
			"ex: partimage -z1 -om save /dev/hda9 /mnt/backup/win95-osr2.partimg.gz\n"
			"ex: partimage imginfo /mnt/backup/debian-potato-2.2.partimg.bz2\n"
			"ex: partimage -a/dev/hda6#/mnt/partimg#vfat -V1440 save /dev/hda12 /mnt/partimg/redhat-6.2.partimg.gz\n\n"
			"Arguments:\n"
			"* <action>:\n"
			"  - save: save the partition datas in an image file\n"
			"  - restore: restore the partition from an image file\n"
			"  - restmbr: restore a MBR of the image file to an hard disk\n"
			"  - imginfo: show informations about the image file\n"
			"  - imgconv: convert an image file to an image with another compression level\n"
			"* <device>: partition to save/restore (example: /dev/hda1)\n"
			"* <image_file>: file where data will be read/written. can be very big\n\n"
			"Options:\n"
			"* -z,  --compress     (image file compression level):\n"
			"  -z0, --compress=0   don't compress: very fast but very big image file\n"
			"  -z1, --compress=1   compress using lzo: fast and small image file (default)\n"
			"  -z2, --compress=2   compress using gzip: fast and small image file (default)\n"
			"  -z3, --compress=3   (compress using bzip2: very slow and very small image file):\n"
			"* -c,  --nocheck      don't check the partition before saving\n"
			"* -o,  --overwrite    overwrite the existing image file without confirmation\n"
			"* -d,  --nodesc       don't ask any description for the image file\n"
			"* -V,  --volume       (split image into multiple volumes files)\n"
			"  -VX, --volume=X     create volumes with a size of X KB\n"
			"* -w,  --waitvol      wait for a confirmation after each volume change\n"
			"* -e,  --erase        erase empty blocks on restore with zero bytes\n"
			"* -m,  --allowmnt     don't fail if the partition is mounted. Dangerous !\n"
			"* -M,  --nombr        don't create a backup of the MBR (Mast Boot Record) in the image file\n"
			"* -h,  --help         show help\n"
			"* -v,  --version      show version\n"
			"* -i,  --compilinfo   show compilation options used\n"
			"* -f,  --finish       (action to do if finished successfully):\n"
			"  -f0, --finish=0     wait: don't make anything\n"
			"  -f1, --finish=1     quit\n"
			"  -f2, --finish=2     reboot (restart the computer):\n"
			"  -f3, --finish=3     halt (power off) the computer\n"
			"* -b,  --batch        batch mode: the GUI won't wait for an user action\n"
			"* -y,  --nosync       don't synchronize the disks at the end of the operation (dangerous)\n"
			"* -sX, --server=X     give partimaged server's ip address\n"
			"* -pX, --port=X       give partimaged server's listening port\n"
			"* -g,  --debug=X      set the debug level to X (default: 1):\n"
			"* -n,  --nossl        disable SSL in network mode\n"
			"* -S,  --simulate     simulation of restoration mode\n"
			"* -aX, --automnt=X    automatic mount with X options. Read the doc for more details\n"
			"* -A   --saveall      save all blocks of the partition, even free ones.\n"
			"* -EX  --encrypt=X    encrypt the image file, and the passord is X\n"
			"* -IX  --workadr=X    give the ip adress of the worker server\n"
			"* -kX  --workport=X   give the port of the worker server\n"),
		   VERSION, isVersionStable(VERSION) ? i18n("stable") : i18n("unstable")); 
  nLen += snprintf(szText, nMaxLen-nLen, "===============================================================================\n");
}
