/***************************************************************************
                          buffer.cpp  -  description
                             -------------------
    begin                : Wed Dec 13 2000
    copyright            : (C) 2000 by Franck Ladurelle
    email                : ladurelf@partimage.org
 ***************************************************************************/
// $Revision: 1.1.1.1 $
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#include <pthread.h>
#include <stdio.h>

#ifdef LEVELSUPPORT_GZIP
#include <zlib.h>
#endif // LEVELSUPPORT_GZIP
#ifdef LEVELSUPPORT_BZIP2
#include <bzlib.h>
#endif // LEVELSUPPORT_BZIP2
#ifdef LEVELSUPPORT_LZO
#include <lzo1x.h>
#endif // LEVELSUPPORT_LZO

#include "buffer.h"
#include "imagefile.h"
#include "misc.h"
#include "exceptions.h"
#include "minmax.h"

// MULTITHREAD
int g_nThreadReadState = 0; // is it the end of the operation
int g_nThreadWriteState = 0; // is it the end of the operation
pthread_t g_threadReadBuffer;
pthread_t g_threadWriteBuffer;
BOOL g_bSpaceOnDisk = true;
int g_nThreadReadError; // manage exceptions from the buffer thread (send them to the main thread)
int g_nThreadWriteError; // manage exceptions from the buffer thread (send them to the main thread)

// for time report only
DWORD g_dwTimeThreadUser = 0;
DWORD g_dwTimeThreadSys = 0;

static char g_cChunkBufRead[MAX_CHUNKSIZE];
static char g_cChunkBufWrite[MAX_CHUNKSIZE];

// =======================================================
int writeChunk(CImageBase *cid, char *cBuffer, DWORD dwChunkSize, DWORD dwCompressAlgo, QWORD qwPos, WORD wKind)
{
  BEGIN;

  if (dwChunkSize == 0)
    RETURN_int(0);
  
  DWORD dwDestLen;
  int nRes;
  char szTemp[2048];
  char cDest[MAX_CHUNKSIZE];
  char *cCurDest;
  DWORD dwTotalData;
  CChunkHeader head;
#ifdef LEVELSUPPORT_LZO
  char cLzoBuffer[LZO1X_1_MEM_COMPRESS];
#endif // LEVELSUPPORT_LZO

  // init
  memset(&head, 0, sizeof(head));

  // compress buffer
  cCurDest = cDest+sizeof(head);
  dwDestLen = sizeof(cDest);
  switch (dwCompressAlgo)
    {
    case COMPRESS_NONE:
      memcpy(cCurDest, cBuffer, dwChunkSize);
      dwDestLen = dwChunkSize;
      break;

#ifdef LEVELSUPPORT_GZIP
    case COMPRESS_GZIP:
      nRes = compress((Bytef*)cCurDest, &dwDestLen, (Bytef*)cBuffer, dwChunkSize);
      if ((nRes != Z_OK) || (!dwDestLen))
	{
	  showDebug(1, "ERROR: GZIP: uncompress()=%d=[%s]\n", nRes, formatGzipError(nRes, szTemp, sizeof(szTemp)));
	  RETURN_int(-1);
	}
      break;
#endif // LEVELSUPPORT_GZIP

#ifdef LEVELSUPPORT_LZO
    case COMPRESS_LZO:
      if (lzo_init() != LZO_E_OK)
	{
	  showDebug(1, "ERROR: LZO: lzo_init() failed\n");
	  RETURN_int(-1);
	}
      nRes = lzo1x_1_compress((BYTE*)cBuffer, dwChunkSize, (BYTE*)cCurDest, (lzo_uint*)&dwDestLen, cLzoBuffer);
      if (nRes != LZO_E_OK)
	{
	  showDebug(1, "ERROR: LZO: lzo1x_1_compress()=%d\n", nRes);
	  RETURN_int(-1);
	}
      break;
#endif // LEVELSUPPORT_LZO

#ifdef LEVELSUPPORT_BZIP2
    case COMPRESS_BZIP2:
      nRes = BZ2_bzBuffToBuffCompress(cCurDest, (unsigned int *)&dwDestLen, cBuffer, dwChunkSize, 8, 1, 30);
      if (nRes != BZ_OK)
	{
	  showDebug(1, "ERROR: BZIP2: BZ2_bzBuffToBuffCompress()=%d=[%s]\n", nRes, formatBzip2Error(nRes, szTemp, sizeof(szTemp)));
	  RETURN_int(-1);
	}
      break;
#endif // LEVELSUPPORT_BZIP2
    }
  
  // fill header
  head.dwCRC = increaseCrc32(0, NULL, 0);
  head.dwCRC = increaseCrc32(head.dwCRC, cCurDest, dwDestLen);
  strncpy(head.cMagic, "CHK", 3);
  head.qwPos = qwPos; 
  head.wKind = wKind;
  head.dwCompress = dwCompressAlgo;
  head.dwChunkCompressedSize = dwDestLen;
  memcpy(cDest, &head, sizeof(head));
  //showDebug(6, "HEADER: dwCompressAlgo=%lu and dwDestLen=%lu and dwCRC=%lu and qwPos=%llu\n", head.dwCompressAlgo, head.dwChunkCompressedSize, head.dwCRC, qwPos);

  showDebug(6, "BUF_HEAD: dwCompressAlgo=%lu and dwDestLen=%lu and dwCRC=%lu and qwPos=%llu and wKind=%u=%s\n", 
	    dwCompressAlgo, dwChunkSize, head.dwCRC, qwPos, wKind, formatKind(wKind, szTemp));

  dwTotalData = dwDestLen + sizeof(head);
  showDebug(4, "HEAD:%d\n", sizeof(head));
  try 
    {
      showDebug(3, "SIZES: %ld %ld\n", sizeof(cDest), dwTotalData);
      cid->write((void*)cDest, dwTotalData/*16400*/, &nRes);
      if (nRes != 0)
	{
	  showDebug(1, "ERROR1: cid->write()=%d\n",nRes);
	  RETURN_int(nRes);
	}
    }
  catch(CExceptions *except)
    {
      showDebug(1, "ERROR2: exception=%d\n",except->GetExcept());
      RETURN_int(nRes);
    }

  RETURN_int(0);
}

// =======================================================
int readChunk(CImageBase *cid, char *cBuffer, DWORD dwMaxBufSize, QWORD qwPos, CSenderHeader *headSend, DWORD *dwCompSize)
{
  BEGIN;

  int nRes;
  char szTemp[2048];
  char cDest[1024*1024];
  CChunkHeader head;
  DWORD dwRes;
  DWORD dwNewCrc;

  // read header
  dwRes = cid->read(&head, sizeof(head));
  if (dwRes != sizeof(head))
    {
      showDebug(1, "EOF: dwRes=%lu and sizeof()=%lu\n", dwRes, (DWORD)sizeof(head));
      RETURN_int(ERR_EOF);
    }

  showDebug(6, "BUF_HEAD: dwCompressAlgo=%lu and dwDestLen=%lu and qwPos=%llu and wKind=%u=%s\n", 
	    head.dwCompress, head.dwChunkCompressedSize, head.qwPos, head.wKind, formatKind(head.wKind, szTemp));

  //showDebug(3, "KINDER=%u\n", head.wKind);

  headSend->dwCompression = head.dwCompress;
  *dwCompSize = head.dwChunkCompressedSize;
  headSend->wKind = head.wKind;

  // check magic
  if ((head.cMagic[0] != 'C') || (head.cMagic[1] != 'H') || (head.cMagic[2] != 'K'))
    {
      showDebug(1, "CHECK-MAGIC error [%3s] [%3s] and errno=%d=%s\n", "CHK", head.cMagic, errno, strerror(errno));
      
      BYTE *a = (BYTE *)&head;
      for (int i=0; i < 20; i++)
	showDebug(1, "byte[%d]=%.2X=%c\n",i,a[i],a[i]);

      RETURN_int(ERR_WRONGMAGIC);
    }

  // check position
  if (head.qwPos != qwPos)
    {
    showDebug(1, "CHECK-POS: old=%llu and new=%llu\n", head.qwPos, qwPos);
      
    /*nRes = g_interface->msgBoxYesNo(i18n("local CRC error"),
    i18n("There was a position error in CCheck. "
    "Data may be damaged. Do you want to continue ?"));
    if (nRes == MSGBOX_NO)*/
    RETURN_int(ERR_CHECK_NUM);
    //THROW(ERR_CHECK_NUM, i, check.qwPos);
    }
  

  dwRes = cid->read((void*)cDest, head.dwChunkCompressedSize);
  if (dwRes != head.dwChunkCompressedSize)
    {
      showDebug(1, "EOF: dwRes=%lu and sizeof()=%lu\n", dwRes, (DWORD)head.dwChunkCompressedSize);
      RETURN_int(ERR_INCOMPLETE);
    }

  // check CRC
  dwNewCrc = increaseCrc32(0, NULL, 0);
  dwNewCrc = increaseCrc32(dwNewCrc, cDest, head.dwChunkCompressedSize);

  if (dwNewCrc != head.dwCRC)
    {
      showDebug(1, "CHECK-CRC: old=%lu and new=%lu\n", head.dwCRC, dwNewCrc);

      /*nRes = g_interface->msgBoxYesNo(i18n("local CRC error"), i18n("There was a CRC error in CCheck. "
								    "Data may be damaged. Do you want to continue ?"));
      if (nRes == MSGBOX_NO)*/
      RETURN_int(ERR_CHECK_CRC);
      //THROW(ERR_CHECK_CRC, dwNewCrc, head.dwCRC);
    }

  // decompress buffer
  headSend->dwDataSize = dwMaxBufSize;
  switch (head.dwCompress)
    {
    case COMPRESS_NONE:
      memcpy(cBuffer, cDest, head.dwChunkCompressedSize);
      headSend->dwDataSize = head.dwChunkCompressedSize;
      break;

#ifdef LEVELSUPPORT_GZIP
    case COMPRESS_GZIP:
      nRes = uncompress((Bytef*)cBuffer, &(headSend->dwDataSize), (Bytef*)cDest, head.dwChunkCompressedSize);
      if ((nRes != Z_OK) || (!headSend->dwDataSize))
	{
	  showDebug(1, "ERROR: GZIP: uncompress()=%d=[%s]\n", nRes, formatGzipError(nRes, szTemp, sizeof(szTemp)));
	  RETURN_int(-1);
	}
      break;
#endif // LEVELSUPPORT_GZIP

#ifdef LEVELSUPPORT_LZO
    case COMPRESS_LZO:
      if (lzo_init() != LZO_E_OK)
	{
	  showDebug(1, "ERROR: LZO: lzo_init() failed\n");
	  RETURN_int(-1);
	}
      nRes = lzo1x_decompress((BYTE*)cDest, head.dwChunkCompressedSize, (BYTE*)cBuffer, (lzo_uint*)&(headSend->dwDataSize), NULL);
      if (nRes != LZO_E_OK)
	{
	  showDebug(1, "ERROR: LZO: lzo1x_decompress()\n", nRes);
	  RETURN_int(-1);
	}
      break;
#endif // LEVELSUPPORT_LZO

#ifdef LEVELSUPPORT_BZIP2
    case COMPRESS_BZIP2:
      nRes = BZ2_bzBuffToBuffDecompress(cBuffer, (unsigned int *)&headSend->dwDataSize, cDest, head.dwChunkCompressedSize, 0, 1);
      if (nRes != BZ_OK)
	{
	  showDebug(1, "ERROR: BZIP2: BZ2_bzBuffToBuffDecompress()=%d=[%s]\n", nRes, formatBzip2Error(nRes, szTemp, sizeof(szTemp)));
	  RETURN_int(-1);
	}
      break;
#endif // LEVELSUPPORT_BZIP2
    }

  //showDebug(6, "HEDD: dwCompressAlgo=%lu and dwCmpSize=%lu and dwFullSize=%lu and dwCRC=%lu and qwPos=%llu\n", 
  //head.dwCompressAlgo, head.dwChunkCompressedSize, *dwDataSize, head.dwCRC, qwPos);

  //showDebug(3, "KINDER2=%u\n", headSend->wKind);

  RETURN_int(0);
}

// =======================================================
void checkSplitting(int nRes)
{
  if (nRes != 0)
    {
      if (nRes == ERR_INCOMPLETE)
	{
	  g_bSpaceOnDisk = false;
	  showDebug(1, "ERRORX: So space left\n");
	}
      else
	{
	  showDebug(1, "ERROR received from cid::write(): %d\n", nRes);
	  g_nThreadWriteError = nRes;
	}
    }
}

// =======================================================
void procWriteBufferToImage(void *ptr) // Thread which will write data to image file
{
  int nRes;
  DWORD dwRes;
  DWORD dwSize;
  DWORD dwChunkSize;
  DWORD dwCompressAlgo;
  DWORD dwCurChunkSize;
  FILE *fStream;
  DWORD dwTemp;
  /*static */QWORD qwPos = 0LL;
  WORD wKind;
  WORD wOldKind;
  char szTemp[4096];
  
  CParam *param;
  CImage *image;
  CImageBase *cid;
  
  param = (CParam *)ptr;
  image = param -> image;
  cid = image -> getImageDisk();
  g_nDebugThreadBufferW = getpid();
  g_nThreadWriteState = THREAD_RUNNING;
  dwCompressAlgo = param -> options.dwCompression;

  BEGIN;

  g_bSpaceOnDisk = true;
  g_nThreadWriteError = 0;

  switch (dwCompressAlgo)
    {
    case COMPRESS_NONE:
      dwChunkSize = CHUNKSIZE_NONE;
      break;

    case COMPRESS_GZIP:
      dwChunkSize = CHUNKSIZE_GZIP;
      break;

    case COMPRESS_LZO:
      dwChunkSize = CHUNKSIZE_LZO;
      break;

    case COMPRESS_BZIP2:
      dwChunkSize = CHUNKSIZE_BZIP2; 
      break;

    default:
      dwChunkSize = 0;
      break;
    }

  showDebug(1, "dwCompressAlgo=%lu and dwChunkSize=%lu\n", dwCompressAlgo, dwChunkSize);
  
  fStream = fdopen(param->nFd, "rb");
  if (!fStream)
    {
      showDebug(1, "ERRORX\n");
      //MBD
    }

  // A) HEADER
  nRes = fread(&wKind, 1, sizeof(wKind), fStream);
  nRes = fread(&dwSize, 1, sizeof(dwSize), fStream);
  showDebug(6, "BUF_RECEIVE: wKind=%u=%s and dwSize=%lu\n", wKind, formatKind(wKind, szTemp), dwSize);
  wOldKind = wKind;

  while (!feof(fStream))
    {
      // 0. ------ init
      dwCurChunkSize = 0;

      // 1. ------ read data
      // while the same compress mode (yes or no)
      while ((wOldKind == wKind) && (!feof(fStream)))
	{
	  while (dwSize > 0) // while data to add for this entry
	    {
	      // add data to current chunk
 	      dwTemp = min(dwSize, dwChunkSize-dwCurChunkSize);
	      dwRes = fread(g_cChunkBufWrite+dwCurChunkSize, 1, dwTemp, fStream);
	      dwCurChunkSize += dwRes;
	      dwSize -= dwRes;

	      // empty chunk if it's full
	      if (dwCurChunkSize >= dwChunkSize) // it can only be equal
		{
		  //showDebug(3, "WRITECHUNK: dwCurChunkSize=%lu and wOldKind=%lu\n",dwCurChunkSize,wOldKind);
		  nRes = writeChunk(cid, g_cChunkBufWrite, dwCurChunkSize, (wKind == KIND_BLOCKS) ? dwCompressAlgo : COMPRESS_NONE, qwPos, wKind);
		  checkSplitting(nRes);
		  qwPos += (QWORD)dwCurChunkSize; //showDebug(3, "add %lu\n", dwCurChunkSize);
		  dwCurChunkSize = 0;
		}
	    }

	  // A) HEADER (read next entry)
	  wOldKind = wKind;
	  nRes = fread(&wKind, 1, sizeof(wKind), fStream);
	  nRes = fread(&dwSize, 1, sizeof(dwSize), fStream);
	  showDebug(6, "BUF_RECEIVE: wKind=%u=%s and dwSize=%lu\n", wKind, formatKind(wKind, szTemp), dwSize);
	}

      // flush chunk
      if (dwCurChunkSize)
	{
	  //showDebug(3, "FLUSH: dwCurChunkSize=%lu and wOldKind=%lu\n",dwCurChunkSize,wOldKind);
	  nRes = writeChunk(cid, g_cChunkBufWrite, dwCurChunkSize, (wOldKind == KIND_BLOCKS) ? dwCompressAlgo : COMPRESS_NONE, qwPos, wOldKind);
	  checkSplitting(nRes);
	  qwPos += (QWORD)dwCurChunkSize; 
	  dwCurChunkSize = 0;
	}

      wOldKind = wKind;
    }

  fclose(fStream);
  //g_nThreadWriteState = THREAD_EOF;

  showDebug(1, "EXIT procWriteBufferFromImage after while and g_nThreadWriteState=%d\n", g_nThreadWriteState);

  RETURN;
}

// =======================================================
#define checkFinished(nFd) \
{ \
  if (g_nThreadReadState == THREAD_ASKEXIT)\
    {\
      showDebug(1, "EXIT procReadBufferFromImage inside while and g_nThreadState=THREAD_ASKEXIT\n");\
      close(nFd);\
      g_nThreadReadState = THREAD_FINISHED;\
      RETURN;\
    }\
}

// =======================================================
void procReadBufferFromImage(void *ptr) // Thread which will read data from image file
{
  DWORD dwRes;
  int nRes;

  CParam *param;
  CImage *image;
  CImageBase *cid;
  int nFd;
  int nRead;
  BOOL bSendChunkHeader;
  DWORD dwCompSize;
  CSenderHeader headSend;
  /*static */QWORD qwPos = 0LL;
 
  param = (CParam *)ptr;
  image = param -> image;
  cid = image -> getImageDisk();
  g_nDebugThreadBufferR = getpid();
  nFd = param->nFd;
  bSendChunkHeader = param->bSendChunkHeader;

  BEGIN;

  g_nThreadReadState = THREAD_RUNNING;
  g_bSpaceOnDisk = true;
  g_nThreadReadError = 0;
  dwRes = 0;
  //qwPos = 0LL;

  do
    {
      checkFinished(nFd);

      nRead = readChunk(cid, g_cChunkBufRead, sizeof(g_cChunkBufRead), qwPos, &headSend, &dwCompSize); 
      qwPos += (QWORD)headSend.dwDataSize;

      checkFinished(nFd);
      
      if (nRead == 0) // if good
	{
	  if (bSendChunkHeader)
	    nRes = write(nFd, &headSend, sizeof(headSend));

	  nRes = write(nFd, g_cChunkBufRead, headSend.dwDataSize);
	  if (nRes == -1)
	    {
	      // MBD
	      showDebug(1, "ERRORX\n");
	    }
	}
      else if (nRead != ERR_EOF) // if error
	{
	  g_nThreadReadError = nRead; // ERR_INCOMPLETE;
	  showDebug(1, "cid->read() returned an error\n");
	}	
    }
  while (nRead == 0);

  close(nFd);
  g_nThreadReadState = THREAD_EOF;

  showDebug(1, "EXIT procReadBufferFromImage after while and g_nThreadReadState=%d->%d\n", g_nThreadReadState, THREAD_EOF);
  RETURN;
}
