/***************************************************************************
      fs_fat.cpp  -  FAT (File Allocation Table DOS) File System Support
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "fs_fat.h"
#include "global.h"
#include "errors.h"
#include "exceptions.h"
#include "misc.h"
#include "imagefile.h"
#include "minmax.h"

// =======================================================
CFatPart::CFatPart(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize): CFSBase(szDevice, fDeviceFile, qwPartSize)
{
  BEGIN;
  RETURN;
}

// =======================================================
CFatPart::~CFatPart()
{
  BEGIN;
  RETURN;
}

// =======================================================
int CFatPart::fsck()
{
  BEGIN;

  try { compareFatCopies(); }
  catch ( CExceptions * excep )
    { throw excep; }

  RETURN_int(0); // success
}

// =======================================================
// -1: error
// false: no differences between FATs
// true: differences between FATs
int CFatPart::compareFatCopies()
{	
  BEGIN;

  int nResult;
  int nRes;
  DWORD dwReservedBytes;
  const DWORD dwBufSize = 256*1024;
  char cCompBuf1[dwBufSize];
  char cCompBuf2[dwBufSize];
  QWORD qwFatSize;
  QWORD qwPtr1, qwPtr2;
  DWORD dwTemp;
  QWORD qwToDo;
  
  if (IGET(cNumberOfFATs) < 2)
    RETURN_int(-1);

  // init
  nResult = 0;
  dwReservedBytes = IGET(wReservedSectorsCount) * IGET(wBytesPerSector);
  if (IGET(dwFileSystem) == FS_FAT16) // if FAT16
    qwFatSize = (QWORD)IGET(dwClustersCount)*2LL;
  else if (IGET(dwFileSystem) == FS_FAT32) // if FAT32
    qwFatSize = (QWORD)IGET(dwClustersCount)*4LL;
  else
    return -1; //THROW(ERR_WRONG_FS);
  
  qwPtr1 = (QWORD)dwReservedBytes;
  qwPtr2 = (QWORD)dwReservedBytes+(IGET(dwSectorsPerFAT)*IGET(wBytesPerSector));
  qwToDo = qwFatSize;

  while (qwToDo > 0)
    {
      dwTemp = (DWORD)min((QWORD)dwBufSize, qwToDo);
      
      nRes = readData(cCompBuf1, qwPtr1, dwTemp);
      if (nRes == -1)
	return -1; //THROW(ERR_READING, (DWORD)qwPtr1, errno);

      nRes = readData(cCompBuf2, qwPtr2, dwTemp);
      if (nRes == -1)
	return -1; //THROW(ERR_READING, (DWORD)qwPtr2, errno);

      if (memcmp(cCompBuf1, cCompBuf2, dwTemp) != 0)
	{
	  showDebug(1, "FAT_COMPARE: differences found\n");
	  return -1; //THROW(ERR_FAT_MISMATCH);
	}

      qwPtr1 += (QWORD)dwTemp;
      qwPtr2 += (QWORD)dwTemp;
      qwToDo -= (QWORD)dwTemp;
    }

  showDebug(1, "FAT_COMPARE: no differences found\n");

  RETURN_int(0);
}

// =======================================================
void CFatPart::printfInformations()
{
  char szText[8192];
  char szFullText[8192];

  getStdInfos(szText, sizeof(szText), false);

  SNPRINTF(szFullText, i18n("%s" // standard infos
			    "Sector Size...................%u\n"
			    "Sectors per cluster...........%u\n"
			    "Reserved Sectors count........%u\n"
			    "Root directory sectors........%lu\n"
			    "FAT tables count..............%u\n"
			    "Total Sectors Count...........%lu\n"
			    "Sectors per FAT...............%lu\n"
			    "Clusters count:...............%lu\n"
			    "Used clusters.................%lu\n"
			    "Free clusters.................%lu\n"
			    "Damaged clusters..............%lu\n"),
	   szText, IGET(wBytesPerSector), IGET(cSectorsPerCluster),
	   IGET(wReservedSectorsCount), IGET(dwRootDirSectors),
	   IGET(cNumberOfFATs),IGET(dwTotalSectorsCount),
	   IGET(dwSectorsPerFAT), IGET(dwClustersCount),
	   IGET(dwUsedClusters), IGET(dwFreeClusters),
	   IGET(dwDamagedClusters));

//  g_interface->msgBoxOk(i18n("FAT informations"), szFullText);
}			

// =======================================================
int CFatPart::readBitmap()
{
   BEGIN;

   DWORD i, j;
   DWORD dwCurSect;
   int nRes;
   DWORD dwReservedBytes;
   WORD wFat16Entry;
   DWORD dwFat32Entry;

   DWORD dwDamagedClusters=0;
   DWORD dwFreeClusters=0;
   DWORD dwUsedClusters=0;

   // init
   m_bitmap.init(SGET(m_header,qwBitmapSize)+1024);

   dwCurSect = 0;
   ISET(dwUsedClusters, 0); 
   ISET(dwFreeClusters, 0);

   // ---- A) the reserved sectors are used
   for (i=0; i < IGET(wReservedSectorsCount); i++,dwCurSect++)
      m_bitmap.setBit(dwCurSect, true);
  
   // ---- B) the FAT tables are on used sectors
   for (j=0; j < IGET(cNumberOfFATs); j++)
      for (i=0; i < IGET(dwSectorsPerFAT); i++,dwCurSect++)
	 m_bitmap.setBit(dwCurSect, true);

   // ---- C) The rootdirectory is on used sectors
   if (IGET(dwRootDirSectors) > 0) // no rootdir sectors on FAT32
      for (i=0; i < IGET(dwRootDirSectors); i++,dwCurSect++)
	 m_bitmap.setBit(dwCurSect, true);
      
   // ---- D) The clusters

   // init
   dwReservedBytes = IGET(wReservedSectorsCount) * IGET(wBytesPerSector);
   ISET(dwDamagedClusters, 0L);
  
   // skip reserved data and two first FAT entries
   if (IGET(dwFileSystem) == FS_FAT16)
      fseek(m_fDeviceFile, dwReservedBytes+(2*2), SEEK_SET);
   else if (IGET(dwFileSystem) == FS_FAT32) 
      fseek(m_fDeviceFile, dwReservedBytes+(2*4), SEEK_SET);
   else
      return -1; //THROW(ERR_WRONG_FS);
  
   //printf ("--- FAT Bitmap: real pos=%llu=%lu\n", ((dwCurSect+2)*512LL), dwCurSect+2);

   // for each FAT entry...
   for (i=0; i < IGET(dwClustersCount); i++)
   {
      // If FAT16
      if (IGET(dwFileSystem) == FS_FAT16)
      {
	 nRes = fread(&wFat16Entry, sizeof(WORD), 1, m_fDeviceFile);
	 if (nRes != 1)
            return -1; //THROW(ERR_READING, i, errno);
	 wFat16Entry = Le16ToCpu(wFat16Entry);
	  
	 if (wFat16Entry == 0xFFF7) // bad FAT16 cluster
	 {
	    dwDamagedClusters++; //m_info.dwDamagedClusters++;
	    for (j=0; j < IGET(cSectorsPerCluster); j++,dwCurSect++) 
	       m_bitmap.setBit(dwCurSect, false);
	 }
	 else if (wFat16Entry == 0x0000) // free
	 {
	    dwFreeClusters++; //m_info.dwFreeClusters++;
	    for (j=0; j < IGET(cSectorsPerCluster); j++,dwCurSect++) 
	       m_bitmap.setBit(dwCurSect, false);
	 }
	 else // used
	 {
	    dwUsedClusters++; //m_info.dwUsedClusters++;
	    for (j=0; j < IGET(cSectorsPerCluster); j++,dwCurSect++) 
	       m_bitmap.setBit(dwCurSect, true);
	 }
      }
      
      // If FAT32
      else if (IGET(dwFileSystem) == FS_FAT32)
      {
	 nRes = fread(&dwFat32Entry, sizeof(DWORD), 1, m_fDeviceFile);
	 if (nRes != 1)
            return -1; //THROW(ERR_READING, (DWORD) 0, errno);
	 dwFat32Entry = Le32ToCpu(dwFat32Entry);
	  
	 if (fat32to28(dwFat32Entry) == 0x0FFFFFF7) // bad FAT32 cluster
	 {	
	    dwDamagedClusters++; //m_info.dwDamagedClusters++;
	    for (j=0; j < IGET(cSectorsPerCluster); j++,dwCurSect++) 
	       m_bitmap.setBit(dwCurSect, false);
	 }
	 else if (fat32to28(dwFat32Entry) == 0x00000000) // free
	 {	
	    dwFreeClusters++; //m_info.dwFreeClusters++;
	    for (j=0; j < IGET(cSectorsPerCluster); j++,dwCurSect++) 
	       m_bitmap.setBit(dwCurSect, false);
	 }
	 else // used
	 {
	    dwUsedClusters++; //m_info.dwUsedClusters++;
	    for (j=0; j < IGET(cSectorsPerCluster); j++,dwCurSect++) 
	       m_bitmap.setBit(dwCurSect, true);
	 }
      }
      
      // If other file system
      else
	 return -1; //THROW(ERR_WRONG_FS);
      
   }

   // mark final sectors as used
   while (dwCurSect < IGET(dwTotalSectorsCount))
   {
      m_bitmap.setBit(dwCurSect, true);
      dwCurSect++;
   }

   ISET(dwDamagedClusters, dwDamagedClusters);
   ISET(dwFreeClusters, dwFreeClusters);
   ISET(dwUsedClusters, dwUsedClusters);  

   // ---- end
   calculateSpaceFromBitmap();

   RETURN_int(0);
}

// =======================================================
int CFatPart::readSuperBlock()
{
   BEGIN;

   BYTE cBoot[MAX_BOOT_SIZE];
   WORD wTemp;
   DWORD dwTemp;
   DWORD i;
   int nRes;
  
   // init
   memset(&m_info, 0, sizeof(CInfoFatHeader));

   // 1. read and print important informations
   nRes = readData(&cBoot, 0, MAX_BOOT_SIZE);
   if (nRes == -1)
   {
      // g_interface->msgBoxError(i18n("Can't read boot sector"));
      return -1; //THROW(ERR_READING, (DWORD) 0, errno);
   }

   // 2. calculate infos
   ISET(wBytesPerSector, Le16ToCpu(*((WORD*)(cBoot+11)))); 
   // memcpy(&m_info.wBytesPerSector, cBoot+11, 2); m_info.wBytesPerSector = Le16ToCpu(m_info.wBytesPerSector);

   ISET(cSectorsPerCluster, *(cBoot+13));
   //memcpy(&m_info.cSectorsPerCluster, cBoot+13, 1);

   ISET(wReservedSectorsCount, Le16ToCpu(*((WORD*)cBoot+14)));
   //memcpy(&m_info.wReservedSectorsCount, cBoot+14, 2); m_info.wReservedSectorsCount = Le16ToCpu(m_info.wReservedSectorsCount);
  
   ISET(cNumberOfFATs, *(cBoot+16));
   //memcpy(&m_info.cNumberOfFATs, cBoot+16, 1);

   ISET(wRootEntriesCount, Le16ToCpu(*((WORD*)cBoot+17)));
   //memcpy(&m_info.wRootEntriesCount, cBoot+17, 2); m_info.wRootEntriesCount = Le16ToCpu(m_info.wRootEntriesCount);
  
   ISET(wFsInfoSector, Le16ToCpu(*((WORD*)cBoot+48)));
   //memcpy(&m_info.wFsInfoSector, cBoot+48, 2); m_info.wFsInfoSector = Le16ToCpu(m_info.wFsInfoSector);

   // check data
   if (IGET(wBytesPerSector) % 512 != 0)
      return -1; //THROW(ERR_WRONG_FS, errno);

   if (IGET(wBytesPerSector) < 512)
      return -1; //THROW(ERR_WRONG_FS, errno);

   if (IGET(cSectorsPerCluster) == 0)
      return -1; //THROW(ERR_WRONG_FS, errno);

   if (IGET(wReservedSectorsCount) ==0) // FIXME
      ;//return -1; //THROW(ERR_WRONG_FS, errno);

   if (IGET(cNumberOfFATs) == 0)
      return -1; //THROW(ERR_WRONG_FS, errno);

   // Label
   /*memcpy(&cTemp, cBoot+38, 1);
     if (cTemp == 0x29) // there is a label
     memcpy(m_header.val.szLabel, cBoot+43, 11);
     else*/
   memset(m_header.val.szLabel, 0, 11);
  
   // Total number of sectors
   memcpy(&wTemp, cBoot+19, 2); wTemp = Le16ToCpu(wTemp);
   memcpy(&dwTemp, cBoot+32, 4); dwTemp = Le32ToCpu(dwTemp);
   if (wTemp)
      ISET(dwTotalSectorsCount, (DWORD)wTemp);
   else
      ISET(dwTotalSectorsCount, dwTemp);
  
   // avoid bug: cannot read last sector
   ISET(dwTotalSectorsCount, IGET(dwTotalSectorsCount)-1); // dwTotalSectorsCount--

   // Count of sectors occupied by one FAT
   memcpy(&wTemp, cBoot+22, 2); wTemp = Le16ToCpu(wTemp);
   memcpy(&dwTemp, cBoot+36, 4); dwTemp = Le32ToCpu(dwTemp);
   if (wTemp)
      ISET(dwSectorsPerFAT, (DWORD)wTemp);
   else
      ISET(dwSectorsPerFAT, dwTemp);		
  
   // calculate the FAT Type with the cluster count (FAT12 or FAT16 or FAT32)
   ISET(dwRootDirSectors, ((IGET(wRootEntriesCount) * 32) + (IGET(wBytesPerSector)-1)) / IGET(wBytesPerSector));
  
   ISET(dwDataSectors, IGET(dwTotalSectorsCount) - (IGET(wReservedSectorsCount) + (IGET(cNumberOfFATs) * IGET(dwSectorsPerFAT)) + IGET(dwRootDirSectors)));
  
   if (IGET(cSectorsPerCluster) == 0)
   {
      // g_interface->msgBoxError(i18n("m_cSectorsPerCluster can't be null. Check the partition to save is a FAT one.") );
      return -1; //THROW(ERR_WRONG_FS);
   }
  
   ISET(dwClustersCount, IGET(dwDataSectors) / IGET(cSectorsPerCluster));
  
   setSuperBlockInfos(false, false, 0, 0);
   if (IGET(dwClustersCount) < 4085)
      ISET(dwFileSystem, FS_FAT12);
   else if (IGET(dwClustersCount) < 65525)
   {
      ISET(dwBytesPerFatEntry, 2);
      ISET(dwFileSystem, FS_FAT16);
   }
   else
   {
      ISET(dwBytesPerFatEntry, 4);
      ISET(dwFileSystem, FS_FAT32);
   }

   // data for the abstract file system
   SPUT(m_header,qwBlockSize, IGET(wBytesPerSector));
   SPUT(m_header,qwBitmapSize, (IGET(dwTotalSectorsCount)+7)/8);
   SPUT(m_header,qwBlocksCount, IGET(dwTotalSectorsCount));

   // FS info sector
   showDebug(1, "FSInfo=%u\n", IGET(wFsInfoSector));
  
   if (IGET(wFsInfoSector))
      nRes = readFSInfo(IGET(wFsInfoSector));
   else
      nRes = -1;

   if ((nRes == -1) && (IGET(dwFileSystem) == FS_FAT32))
   {
      showDebug(1, "Can't read FSInfo sector\n");
      // g_interface->msgBoxError(i18n("Can't read FSInfo sector"));
      setSuperBlockInfos(false, false, 0, 0);
   }

   // ------------- get the real label ---------------

   if (IGET(dwFileSystem) == FS_FAT16)
   {
      QWORD qwPos;
      CDirEntry dir;
      DWORD dwPos;

      dwPos = IGET(wReservedSectorsCount)+(IGET(cNumberOfFATs)*IGET(dwSectorsPerFAT));
      //ISET(dwRootDirFirst512Sector, dwPos);
      //qw512SectorOffset = dwPos * 512;
      qwPos = (QWORD)(dwPos * 512);
      //printf ("  FAT16: wPos = sector %lu ", (DWORD)dwPos);
   
      //printf ("final pos=%llu=%llu*512\n", qwPos, (qwPos/512LL));
      fseek(m_fDeviceFile, qwPos, SEEK_SET); // TODO: bigFseek

      for (i=0; i < 100; i++)
      {
	 fread(&dir, sizeof(dir), 1, m_fDeviceFile);
	 
	 //printf ("---> name [%s]\n", dir.szName);
	 /*if ((dir.cAttr & ATTR_VOLUME) != 0)
	 {
	    printf ("---> volume [%s]\n", dir.szName);
	 }*/
      }
   }
   else if (IGET(dwFileSystem) == FS_FAT32)
   {
      memcpy(m_header.val.szLabel, cBoot+0x47, 11);
   }

   RETURN_int(0);
}

// =======================================================
int CFatPart::readFSInfo(WORD wFsInfoSector)
{
  BEGIN;

  QWORD qwFreeSectors;
  QWORD qwUsedSectors;
  CFatFsInfoSector sectFsInfo;
  int nRes;

  nRes = readData(&sectFsInfo, IGET(wFsInfoSector)*IGET(wBytesPerSector), sizeof(sectFsInfo));
  if (nRes == -1)
    {
      showDebug(1, "Can't read FSInfo sector\n");
      return -1;
    }

  if (sectFsInfo.dwMagic != CpuToLe32(0x41615252))
    {
      showDebug(1, "FSInfo sector: bad magic\n");
      return -1;
    }

  
  qwFreeSectors = IGET(cSectorsPerCluster) * Le32ToCpu(sectFsInfo.dwFreeClusters);
  qwUsedSectors =  IGET(dwTotalSectorsCount) - qwFreeSectors;

  setSuperBlockInfos(false, false, qwUsedSectors * ((QWORD)IGET(wBytesPerSector)), qwFreeSectors * ((QWORD)IGET(wBytesPerSector)), true);

  RETURN_int(0);
}

