/***************************************************************************
           fs_ufs.cpp  - UFS (Unix File System) File System Support
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
  + Be aware: (the block for partimage) == (a frag for FFS)
    For example, (FFS blocks are 8192 bytes) while (FFS frags == pi blocks == 1024 bytes)
  + Mounting FFS under Linux:
    mount -t ufs -o ro,ufstype=44bsd /dev/xxx /mnt/point
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "fs_ufs.h"
#include "global.h"
#include "misc.h"
#include "errors.h"
#include "exceptions.h"
#include "imagefile.h"

// =======================================================
CUfsPart::CUfsPart(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize): CFSBase(szDevice, fDeviceFile, qwPartSize)
{
  BEGIN;
  RETURN;
}

// =======================================================
CUfsPart::~CUfsPart()
{
  BEGIN;
  RETURN;
}

// =======================================================
int CUfsPart::fsck()
{
  BEGIN;
  RETURN_int(0);
}

// =======================================================
int CUfsPart::readSuperBlock()
{
  BEGIN;

  ffsSuperBlock sb;
  int nRes;
  
  // init
  memset(&m_info, 0, sizeof(CInfoUfsHeader));

  // 0. read super block
  nRes = readData(&sb, SBOFF, sizeof(sb));
  if (nRes == -1)
    return -1; //THROW(ERR_READING);

  //localDebug("sizeof(sb)=%d\n", sizeof(sb));
  //localDebug("sb.fs_sblkno=%llu\n",(QWORD)sb.fs_sblkno);
  //localDebug("sb.fs_cblkno=%llu // offset of cyl-block in filesys\n", (QWORD)sb.fs_cblkno);
  //localDebug("sb.fs_iblkno=%llu // offset of inode-blocks in filesys\n", (QWORD)sb.fs_iblkno);
  //localDebug("sb.fs_dblkno=%llu // offset of first data after cg\n", (QWORD)sb.fs_dblkno);
  //localDebug("sb.fs_cgoffset=%llu // cylinder group offset in cylinder\n", (QWORD)sb.fs_cgoffset);
  //localDebug("sb.fs_size=%llu // number of blocks (frag ?) in fs\n", (QWORD)sb.fs_size);
  //localDebug("sb.fs_dsize=%llu // number of data blocks (frag ?) in fs\n", (QWORD)sb.fs_dsize);
  //localDebug("sb.fs_ncg=%llu // number of cylinder groups\n", (QWORD)sb.fs_ncg);
  //localDebug("sb.fs_bsize=%llu // size of basic blocks in fs\n", (QWORD)sb.fs_bsize);
  //localDebug("sb.fs_fsize=%llu // size of frag blocks in fs\n", (QWORD)sb.fs_fsize);
  //localDebug("sb.fs_frag=%llu // number of frags in a block in fs\n", (QWORD)sb.fs_frag);
  //localDebug("sb.fs_cssize=%llu\n", (QWORD)sb.fs_cssize);
  //localDebug("sb.fs_fshift=%llu\n", (QWORD)sb.fs_fshift);
  //localDebug("sb.fs_csaddr=%llu // blk addr of cyl grp summary area\n", (QWORD)sb.fs_csaddr);
  //localDebug("sb.fs_maxbpg=%llu // max number of blks per cyl group\n", (QWORD)sb.fs_maxbpg);
  //localDebug("sb.fs_cgsize=%llu // cylinder group size (in basic blocks)\n", (QWORD)sb.fs_cgsize);
  //localDebug("\n");
  ////localDebug("sb.=%llu\n", (QWORD)sb.);
  ////localDebug("sb.=%llu\n", (QWORD)sb.);

  SPUT(m_header,qwBlockSize, sb.fs_fsize);
  SPUT(m_header,qwBlocksCount, sb.fs_size);
  SPUT(m_header,qwBitmapSize, (SGET(m_header,qwBlocksCount)+7)/8); 

  ISET(dwCylinderGroupsCount, sb.fs_ncg);
  ISET(fs_fpg, sb.fs_fpg);
  ISET(fs_cgoffset, sb.fs_cgoffset);
  ISET(fs_cgmask, sb.fs_cgmask);
  ISET(fs_cblkno, sb.fs_cblkno);
  ISET(dwFragsPerBlock, sb.fs_frag);
  ISET(dwCylinderGroupSize, sb.fs_cgsize);
  ISET(dwBasicBlockSize, sb.fs_bsize);
  ISET(qwDataFrags, sb.fs_dsize);

  RETURN_int(0);  
}

// =======================================================
int CUfsPart::readBitmap(/*COptions *options*/)
{
  BEGIN;

  cylinderGroupHeader *cgp;
  BYTE cBuffer[65536];
  QWORD qwStatsFreeFrags;
  QWORD qwStatsUsedFrags;  
  DWORD dwCurBmpSize;
  DWORD dwCgBmpSize;
  QWORD qwUgcMin;
  DWORD i, j;
  int nRes;

  // init bitmap -> all blocks marked as used (1)
  m_bitmap.init(SGET(m_header,qwBitmapSize)+1024);
  for (i=0; i < SGET(m_header,qwBlocksCount); i++)
    m_bitmap.setBit(i, true);

  cgp = new cylinderGroupHeader[IGET(dwCylinderGroupsCount)];
  if (!cgp)
    return -1; //THROW(ERR_NOMEM);

  dwCurBmpSize = 0;
  QWORD qwCurStart = 0;
  qwStatsFreeFrags = 0;

  // process all cylinder groups
  for (i=0; i < IGET(dwCylinderGroupsCount); i++)
    {
      qwUgcMin = ufs_cgcmin(IGET(fs_fpg) , IGET(fs_cgoffset), IGET(fs_cgmask),IGET(fs_cblkno),i);
      //localDebug("\nCGC[%d] stored in block (frag ?) %llu\n", i, qwUgcMin);
      
      nRes = readData(&(cgp[i]), qwUgcMin*SGET(m_header,qwBlockSize), sizeof(cylinderGroupHeader));
      if (nRes == -1)
	return -1; //THROW(ERR_ERRNO, errno);
      
      if ((DWORD)cgp[i].cg_magic != (DWORD)CG_MAGIC)
	{
	  //localDebug("ERROR: %d not a cylinder group header\n", i);
	  return -1; //THROW(ERR_ERRNO, errno);
	}

      //localDebug("   from %llu to %llu\n", (QWORD)qwCurStart, (QWORD)qwCurStart+cgp[i].cg_ndblk);

      ////localDebug("cgh[%d].cg_magic=%.8x (CG_MAGIC=%.8x)\n",i,(DWORD)cgp[i].cg_magic, (DWORD)CG_MAGIC);
      ////localDebug("cgh[%d].cg_cgx=%lu // CG number\n",i,(DWORD)cgp[i].cg_cgx);
      //localDebug("cgh[%d].cg_ndblk=%lu // blocks (frag) in this CG\n",i,(DWORD)cgp[i].cg_ndblk);
      //localDebug("cgh[%d].cg_freeoff=%lu // free block map\n",i,(DWORD)cgp[i].cg_freeoff);      
      //localDebug("cgh[%d].cg_cs.cs_ndir=%lu // stats: number of directories\n",i,(DWORD)cgp[i].cg_cs.cs_ndir);      
      //localDebug("cgh[%d].cg_cs.cs_nbfree=%lu // stats: number of free blocks\n",i,(DWORD)cgp[i].cg_cs.cs_nbfree);      
      //localDebug("cgh[%d].cg_cs.cs_nifree=%lu // stats: number of free inodes\n",i,(DWORD)cgp[i].cg_cs.cs_nifree);      
      //localDebug("cgh[%d].cs_cs.cs_nffree=%lu // stats: number of free frags\n",i,(DWORD)cgp[i].cg_cs.cs_nffree);      
      //localDebug("cgh[%d].cg_rotor=%lu // position of last used block\n",i,(DWORD)cgp[i].cg_rotor);      
      //localDebug("cgh[%d].cg_frotor=%lu // position of last used frag\n",i,(DWORD)cgp[i].cg_frotor);      
      //localDebug("cgh[%d].cg_irotor=%lu // position of last used inode \n",i,(DWORD)cgp[i].cg_irotor);      
      //localDebug("cgh[%d].cg_nclusterblks=%lu // number of clusters this cg\n",i,(DWORD)cgp[i].cg_nclusterblks);      
      ////localDebug("cgh[%d].=%lu // \n",i,(DWORD)cgp[i].);      
      ////localDebug("cgh[%d].=%lu // \n",i,(DWORD)cgp[i].);      
      
      qwStatsFreeFrags += (IGET(dwFragsPerBlock) * cgp[i].cg_cs.cs_nbfree) + cgp[i].cg_cs.cs_nffree;

      dwCgBmpSize = (((DWORD)cgp[i].cg_ndblk)+7)/8;
      if (dwCgBmpSize > sizeof(cBuffer))
	{
	  showDebug(1, "FFS: buffer too small: have %lu, need %lu\n", sizeof(cBuffer), dwCgBmpSize);
	  delete []cgp;
	  return -1; //THROW(ERR_ERRNO);
	}

      //localDebug("dwCgBmpSize=%lu\n",(DWORD)dwCgBmpSize);

      // copy and convert bitmap
      nRes = readData(cBuffer, ((qwUgcMin*SGET(m_header,qwBlockSize)) + (DWORD)cgp[i].cg_freeoff), dwCgBmpSize);
      for (j=0; j < dwCgBmpSize; j++)
	cBuffer[j] = ~cBuffer[j];
      memcpy(m_bitmap.data()+dwCurBmpSize, cBuffer, dwCgBmpSize);

      dwCurBmpSize += dwCgBmpSize;
      qwCurStart += (QWORD)cgp[i].cg_ndblk;

    }

  qwStatsUsedFrags = SGET(m_header,qwBlocksCount) - qwStatsFreeFrags;
  setSuperBlockInfos(true, true, qwStatsUsedFrags*SGET(m_header,qwBlockSize), qwStatsFreeFrags*SGET(m_header,qwBlockSize));

  delete []cgp;

  // fill informations about free/used clusters count
  calculateSpaceFromBitmap();

  RETURN_int(0);
}			

// =======================================================
void CUfsPart::printfInformations()
{
  char szFullText[8192];
  char szText[8192];
  char szTemp1[1024];
  
  getStdInfos(szText, sizeof(szText), true);
  
  SNPRINTF(szFullText, i18n("%s" // standard infos
			    "Cylinder groups count:........%lu\n"
			    "Cylinder group size:..........%s\n"
			    "Basic blocks per CG:..........%lu\n"
			    "Basic block size:.............%lu\n"
			    "Data frags count:.............%llu\n"),
	   szText, IGET(dwCylinderGroupsCount), 
	   formatSize(IGET(dwCylinderGroupSize)*IGET(dwBasicBlockSize),szTemp1),
	   IGET(dwCylinderGroupSize), IGET(dwBasicBlockSize), IGET(qwDataFrags));
    
  //g_interface->msgBoxOk(i18n("FFS informations"), szFullText);
}			

