/***************************************************************************
                          image_disk.cpp  -  description
                             -------------------
    begin                : Mon Aug 14 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#ifndef _GNU_SOURCE
  #define _GNU_SOURCE
#endif // _GNU_SOURCE

#ifndef _FILE_OFFSET_BITS
  #define _FILE_OFFSET_BITS 64
#endif //_FILE_OFFSET_BITS

#ifdef HAVE_SYS_PARAM_H
  #include <sys/param.h>
#endif
#ifdef HAVE_SYS_STATFS_H
  #include <sys/statfs.h>
#endif
#ifdef HAVE_SYS_MOUNT_H
  #include <sys/mount.h>
#endif

#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/file.h>

#include "image_disk.h"
#include "errors.h"
#include "common.h"
#include "buffer.h"
#include "global.h"

#define SPACEFILE_SIZE (10*1024*1024) // 10 MB

#ifndef HAVE_FSTAT64
  #warning "fstat32 used"
#endif
#ifndef HAVE_FSTATFS64
  #warning "fstatfs32 used"
#endif
#ifndef HAVE_OPEN64
  #warning "open32 used"
#endif

// =======================================================
CImageDisk::CImageDisk(COptions * options):CImageBase() // [Main-Thread]
{
  m_bIsOpened = false;
  m_qwTotal = 0LL;
  m_qwVolSize = 0LL;
  memcpy(&m_options, options, sizeof(COptions));
  memset(m_szSpaceFilename, 0, sizeof(m_szSpaceFilename));
  showDebug(3, "compress=%ld\n", options->dwCompression);
}

// =======================================================
CImageDisk::~CImageDisk() // [Main-Thread]
{
  BEGIN;
  showDebug(1, "total written: %llu\n", m_qwTotal);
  unlockImageFile();
  RETURN;
}

// =======================================================
QWORD CImageDisk::getPartitionSize(const char *szDevice) // [Main-Thread]
{
  QWORD qwSize;
  int fdDevice;

  showDebug(1, "begin\n");
  qwSize = 0;
	
#ifdef HAVE_OPEN64
  fdDevice = open64(szDevice, O_RDONLY);
#else 
  fdDevice = open(szDevice, O_RDONLY);
#endif
  if (fdDevice < 0)
    THROW(ERR_ERRNO, errno);

#ifdef OS_LINUX
  if (!ioctl(fdDevice, BLKGETSIZE, &lSize))
    {
      qwSize = ((QWORD) 512) * ((QWORD) lSize);
    }
  else
    THROW(ERR_ERRNO, errno);
#endif

  if (::close(fdDevice))
    THROW(ERR_ERRNO, errno);

  return qwSize;
}

// =======================================================
void CImageDisk::getFileSize(QWORD *qwFileSize, char *szFilepath) // [Main-Thread]
{
#ifdef HAVE_FSTAT64
  struct stat64 fStat;
#else
  struct stat fStat;
#endif
  //struct stat st;
  int nRes;

#ifdef HAVE_FSTAT64
  nRes = stat64(szFilepath, &fStat);
#else
  nRes = stat(szFilepath, &fStat);
#endif

  showDebug(1, "getFileSize() -> nRes=%d\n", nRes);

  if (nRes != -1)
    *qwFileSize = (QWORD)fStat.st_size; // ::getFileSize(szFilepath);
  else
    *qwFileSize = (QWORD)0LL; // ::getFileSize(szFilepath);
}

// =======================================================
void CImageDisk::get_qwCurrentVolumeSize(QWORD *qwSize) // [Main-Thread]
{
  //struct stat fStat;

  if (m_bIsOpened == false)
    {
      showDebug(3, "Failed: m_bIsOpened == false\n");
      THROW( ERR_OPENED);
    }

  errno = 0;
  *qwSize = m_qwVolSize;
}

// =======================================================
// may throw incomplete on error
void CImageDisk::write(void *vBuf, DWORD dwLength, int *nResult) // [Buffer-Thread]
{
  DWORD dwRes, dwFirstWritten;
  QWORD qwVolumeSize;
  int nRes;
  char *cBuf;

  showDebug(3, "called for %ld\n", dwLength);
  *nResult = 0;

  if (get_bIsOpened() == false)
    {
      showDebug(3, "file not opened\n");
      THROW(ERR_OPENED);
    }
  
  //*dwResult = 0;
  m_qwTotal += (QWORD) dwLength;
  m_qwVolSize += (QWORD) dwLength;
  cBuf = (char*) vBuf;

  dwRes = ::write(m_nImageFd, cBuf, dwLength); //dwRes = fwrite(cBuf, 1, dwLength, m_fImageFile); 
  //*dwResult += dwRes;

  //if (dwRes != 1) // out of space
  if (dwRes != dwLength) // out of space
    {
      showDebug(1, "ERROR: dwRes != dwLength (dwRes=%lu and dwLen=%lu)\n",dwRes,dwLength);

      dwFirstWritten = dwRes;
      
      // 1. free a little space
      nRes = destroySpaceFile(); 
      if (nRes == -1)
	THROW(ERR_DESTROYSPACEFILE);

      // 2. write the end of data
      dwRes = ::write(m_nImageFd, cBuf+dwFirstWritten, dwLength-dwFirstWritten); //dwRes = fwrite(cBuf+dwFirstWritten, 1, dwLength-dwFirstWritten, m_fImageFile);
      //*dwResult += dwRes;
      if (dwRes != dwLength-dwFirstWritten) // out of space
	{
	  *nResult = ERR_INCOMPLETEFINISH;
	  showDebug(3, "ERROR: ERR_INCOMPLETEFINISH (dwRes != dwLength-dwFirstWritten) error %d = %s [dwRes=%lu and dwLength-dwFirstWritten=%lu]\n", 
		    errno, strerror(errno), dwRes, dwLength-dwFirstWritten);
	  newtWinMessage("a", "b", "ERROR: ERR_INCOMPLETEFINISH (dwRes != dwLength-dwFirstWritten) error %d = %s [dwRes=%lu and dwLength-dwFirstWritten=%lu]\n", 
			 errno, strerror(errno), dwRes, dwLength-dwFirstWritten);
	}
      else // no error
	{
	  *nResult = ERR_INCOMPLETE;
	  showDebug(3, "ERROR: ERR_INCOMPLETE. error = %d = %s\n", errno, strerror(errno));
	}
    }

  // check for SPLIT_SIZE
  get_qwCurrentVolumeSize(&qwVolumeSize);
  //if ( (m_options.dwSplitMode == SPLIT_SIZE) && (/*get_qwCurrentVolumeSize()*/qwVolumeSize /*+ dwLength*/ + ENDVOL_MARGIN > m_options.qwSplitSize))
  if ( (m_options.qwSplitSize) && (/*get_qwCurrentVolumeSize()*/qwVolumeSize /*+ dwLength*/ + ENDVOL_MARGIN > m_options.qwSplitSize))
    {
      *nResult = ERR_SPLITSIZE;
    }

  showDebug(3, "m_qwTotal = %lld\n", m_qwTotal);
}

// =======================================================
// may raise ERR_OPENED, ERR_EOF
DWORD CImageDisk::read(void *buf, DWORD dwLength) // [Buffer-Thread]
{
  if (m_bIsOpened == false)
    THROW(ERR_OPENED);

  DWORD dwRes;
  errno = 0;
  
  // ****************************************

  //clearerr(m_fImageFile);
  dwRes = ::read(m_nImageFd, buf, dwLength); //dwRes = fread(buf, 1, dwLength, m_fImageFile);

  if (dwRes < dwLength)
    {
      showDebug(1, "EOF: dwRes=%lu and dwLength=%lu\n", dwRes, dwLength);
    }

  m_qwTotal += (QWORD) dwRes; //dwLength;
  //showDebug(1, "m_qwTotal = %lld\n", m_qwTotal);
  return dwRes; // important: must return dwRes even if EOF reached
}

// =======================================================
void CImageDisk::unlockImageFile() // [Main-Thread]
{
#ifndef DEVEL_SUPPORT
  flock(m_nImageFd, LOCK_UN);
  showDebug(4, "UNLOCK file %s\n", m_szImageFilename);
#endif // DEVEL_SUPPORT
}

// =======================================================
// may raise an errno exception
void CImageDisk::close() // [Main-Thread]
{
  BEGIN;
 
  showDebug(1, "close -> %lld\n", m_qwTotal); 
  if (m_bIsOpened == false)
    RETURN;
  
  unlockImageFile();

  int nRes = 0;
  errno = 0;

  destroySpaceFile();

  nRes = ::close(m_nImageFd); //fclose(m_fImageFile);	
  if (nRes)
    THROW(ERR_ERRNO, errno);

  m_bIsOpened = false;
  
  RETURN;
}

// =======================================================
int CImageDisk::getCompressionLevelForImage(char *szFilename) // [Main-Thread]
{
  BEGIN;

  FILE *fImageFile;
  DWORD dwRes1, dwRes2; //, dwRes3;
  CVolumeHeader headVolume;
  CChunkHeader headChunk;
  
  fImageFile = fopen(szFilename, "rb");
  if (fImageFile == NULL)
    RETURN_int(-1);
  
  // skip first header
  dwRes1 = fread(&headChunk, 1, sizeof(headChunk), fImageFile);
  dwRes2 = fread(&headVolume, 1, sizeof(headVolume), fImageFile);
  fclose(fImageFile);	

  if ((strncmp(headChunk.cMagic, "CHK", 3) != 0) || (strncmp(headVolume.val.szMagicString, MAGIC_BEGIN_VOLUME, strlen(MAGIC_BEGIN_VOLUME)) != 0) 
      || (headChunk.wKind != KIND_VOLHEAD))
    {
      showDebug(1, "ERROR: Read error\n");
      RETURN_int(-1);
    }

  //dwRes3 = fread(&headMain, 1, sizeof(headMain), fImageFile);
  if ((dwRes1 != sizeof(headChunk)) || (dwRes2 != sizeof(headVolume))/* || (dwRes3 != sizeof(headMain))*/)
    {
      showDebug(1, "ERROR: Read error\n");
      RETURN_int(-1);
    }

  // check this is a valid image file
  if (strncmp(headVolume.val.szMagicString, MAGIC_BEGIN_VOLUME, strlen(MAGIC_BEGIN_VOLUME)) != 0)
    {
      showDebug(1, "ERROR: Bad magic\n");
      RETURN_int(-1);
    }

  RETURN_int(headVolume.val.dwCompress); // success
}

// =======================================================
// may throw ERR_OPENED, ERR_NOTFOUND, errno, comp
void CImageDisk::openReading() // [Main-Thread]
{
  //int nRes;
#ifdef HAVE_FSTAT64
  struct stat64 fStat;
#else
  struct stat fStat;
#endif

  BEGIN;
  showDebug(1, "ALREADY=%llu\n", m_qwTotal);
  if (m_bIsOpened == true)
    THROW(ERR_OPENED);

  // check the file exists
  errno = 0;
  if (access(m_szImageFilename, F_OK) == -1) // if the file doesn't exists
    THROW(ERR_NOTFOUND);
  
  showDebug(1, "openReading[%s]\n", m_szImageFilename);

#ifdef OS_LINUX
  #ifdef HAVE_OPEN64
    m_nImageFd = open64(m_szImageFilename, O_RDONLY | O_LARGEFILE | O_NOFOLLOW);
  #else
    m_nImageFd = open(m_szImageFilename, O_RDONLY | O_LARGEFILE | O_NOFOLLOW);
  #endif
#endif
#ifdef OS_FBSD
  m_nImageFd = open(m_szImageFilename, O_RDONLY | O_NOFOLLOW);
#endif
  if (m_nImageFd == -1)
    THROW(ERR_ERRNO, errno);
  else 
    showDebug(1, "preopen successfull for %s\n", m_szImageFilename);

  // ---- avoid RACE CONDITIONS security problems:
  errno = 0;
#ifdef HAVE_FSTAT64
  if (fstat64(m_nImageFd, &fStat) == -1)
#else
  if (fstat(m_nImageFd, &fStat) == -1)
#endif
    {
      showDebug(1, "fstat() returned -1 and errno=%d=[%s]\n", errno, strerror(errno));
      THROW(ERR_ERRNO, errno);
    }

  if (!S_ISREG(fStat.st_mode)) // not a regular file
    THROW(ERR_NOTAREGULARFILE, m_szImageFilename); //BIGH
  
  // ---- lock the image file
  if (lockImage(LOCK_SH, m_nImageFd, m_szImageFilename) == -1)
    THROW(ERR_LOCKED, m_szImageFilename);
  
  showDebug(1, "%s opened ok\n", m_szImageFilename);
	
  m_bIsOpened = true;
}	

// =======================================================
int CImageDisk::destroySpaceFile() // [Main-Thread]
{
  int nRes;

  showDebug(1, "DESTROY SPACE FILE [%s]\n", m_szSpaceFilename);

  //newtWinMessage("before","b","erase %s", m_szSpaceFilename);
  nRes = unlink(m_szSpaceFilename);
  //newtWinMessage("after","b","erase %s", m_szSpaceFilename);
  return nRes;
}

// =======================================================
int CImageDisk::createSpaceFile() // [Main-Thread]
{
  char cBuffer[1024];
  //int nSpaceFd;
  FILE *fStream;
  DWORD dwFileSize;
  int nRes;
  WORD i;

  showDebug(1, "CREATE SPACE FILE [%s]\n", m_szSpaceFilename);

  // write random numbers
  for (i=0; i < sizeof(cBuffer); i++)
    cBuffer[i] = (i ^ 94) % 256;
  
  // open file descriptor ICI
  fStream = openFileDescriptorSecure(m_szSpaceFilename, "a", O_WRONLY | 
     O_CREAT | O_NOFOLLOW | O_TRUNC, S_IREAD | S_IWRITE);

  if (!fStream)
    return -1;

  dwFileSize = 0;
  while (dwFileSize < SPACEFILE_SIZE)
    {
      //nRes = ::write(nSpaceFd, cBuffer, sizeof(cBuffer));
      nRes = fwrite(cBuffer, sizeof(cBuffer), 1, fStream);
      if (nRes != 1)
	{
	  fclose(fStream);
	  destroySpaceFile();
	  return -1;
	}
      dwFileSize += sizeof(cBuffer);
    }

  nRes = fclose(fStream);
  if (nRes == -1)
    return -1;

  return 0; // success
}

// =======================================================
// may raise ERR_OPENED, ERR_EXIST, errno
void CImageDisk::openWriting() // [Main-Thread]
{
#ifdef HAVE_FSTAT64
  struct stat64 fStat;
#else
  struct stat fStat;
#endif
  int flags;
  int nRes;

  BEGIN;

  showDebug(1, "ALREADY=%lld\n", m_qwTotal);

  if (m_bIsOpened == true)
    THROW(ERR_OPENED);
  
  snprintf(m_szSpaceFilename, MAXPATHLEN, "%s_space", m_szImageFilename);
  errno = 0;
  nRes = createSpaceFile();
  if (nRes == -1)
    {
      if (errno == ENOSPC)
	{
	  showDebug(1, "No space left -> cannot create file [%s]\n",
             m_szSpaceFilename);
	  THROW(ERR_CREATESPACEFILENOSPC, m_szSpaceFilename);
	}
      else
	{
	  showDebug(1, "Cannot create SpaceFile[%s]\n", m_szSpaceFilename); 
	  THROW(ERR_CREATESPACEFILE, m_szSpaceFilename);
	}
    }

  errno = 0;
  flags = O_CREAT | O_WRONLY | O_NOFOLLOW;
#ifdef OS_LINUX
  flags |= O_LARGEFILE;
#endif
  if (m_options.bOverwrite != true)
    flags |= O_EXCL;

#ifdef HAVE_OPEN64
  m_nImageFd = open64(m_szImageFilename, flags, S_IREAD | S_IWRITE);
#else
  m_nImageFd = open(m_szImageFilename, flags, S_IREAD | S_IWRITE);
#endif
  if (m_nImageFd == -1)
    if (errno == EEXIST)
      THROW(ERR_EXIST);
    else
      THROW(ERR_ERRNO, errno);

  // ---- avoid RACE CONDITIONS security problems:
#ifdef HAVE_FSTAT64
  if (fstat64(m_nImageFd, &fStat) == -1)
#else
  if (fstat(m_nImageFd, &fStat) == -1)
#endif
    THROW(ERR_ERRNO, errno);
  
  if (!S_ISREG(fStat.st_mode)) // not a regular file
    THROW(ERR_NOTAREGULARFILE, m_szImageFilename);

  // lock the image file
  if (lockImage(LOCK_EX, m_nImageFd, m_szImageFilename) == -1)
    THROW(ERR_LOCKED, m_szImageFilename);

  // truncate the file (now, we can because it's not a symlink)
  if (ftruncate(m_nImageFd, 0L) == -1)
    THROW(ERR_ERRNO, errno);

  m_bIsOpened = true;
  m_qwVolSize = 0LL;

  showDebug(1, "end of CImageDisk::openWriting(), return = ok\n");
}

// =======================================================
int CImageDisk::lockImage(int flag, int nFd, char *szFilename) // [Main-Thread]
{
#ifndef DEVEL_SUPPORT
  int nRes;

  errno = 0;

  nRes = flock(m_nImageFd, flag | LOCK_NB);
  if (nRes == -1)
    RETURN_int(-1);
  showDebug(4, "LOCK file %s\n", m_szImageFilename);
#endif // DEVEL_SUPPORT

  return 0;
}

// =======================================================
// may throw errno
#ifdef HAVE_SYS_STATFS_H
void CImageDisk::getFreeSpace(QWORD *qwAvailDiskSpace) // [Main-Thread]
{
  int nRes;
#ifdef HAVE_FSTATFS64
  struct statfs64 fsInfo;
#else
  struct statfs fsInfo;
#endif

  *qwAvailDiskSpace = 0LL;

#ifdef HAVE_FSTATFS64
  nRes = fstatfs64(m_nImageFd, &fsInfo);
#else
  nRes = fstatfs(m_nImageFd, &fsInfo);
#endif
  if (nRes == -1) // error
    THROW(ERR_ERRNO, errno);
	
  *qwAvailDiskSpace = ((QWORD) fsInfo.f_bavail) * ((QWORD)fsInfo.f_bsize);
}
#else //ifdef HAVE_SYS_STATFS_H
void CImageDisk::getFreeSpace(QWORD *qwAvailDiskSpace)
{ 
  #warning "getFreeSpace always return 0"
  *qwAvailDiskSpace =  0LL;
}
#endif

// =======================================================
BOOL CImageDisk::doesFileExists(char *szPath) // [Main-Thread]
{
  return ::doesFileExists2(szPath);
}

// =======================================================
#ifdef HAVE_SYS_STATFS_H
void CImageDisk::getDiskFreeSpaceForFile(QWORD *qwAvailDiskSpace, char *szFilepath) // [Main-Thread]
{
  //::getDiskFreeSpaceForFile(qwFreeSpace, szFilepath);
  int nRes;
  struct statfs fsInfo;
  
  *qwAvailDiskSpace = 0LL;
  
  errno = 0;
  nRes = statfs(szFilepath, &fsInfo);
  if (nRes == -1) // error
    {
      if (errno == ENOENT) // file doesn't exists
	{	
	  // create an empty file
	  FILE *fFile;
	  fFile = fopen(szFilepath, "w");
	  if (fFile == NULL)
	    RETURN;
	  fclose(fFile);
	  getDiskFreeSpaceForFile(qwAvailDiskSpace, szFilepath);
	  unlink(szFilepath);
	  RETURN;;
	}
      else // another error
	{
	  RETURN;
	}
      
    }	
	
  *qwAvailDiskSpace = ((QWORD) fsInfo.f_bavail) * ((QWORD)fsInfo.f_bsize);
  
  RETURN;
}
#else // HAVE_SYS_STATFS_H
void CImageDisk::getDiskFreeSpaceForFile(QWORD *qwAvailDiskSpace, char *szFilepath) // [Main-Thread]
{
  #warning "getDiskFreeSpaceForFile always return 5 MB"
  *qwAvailDiskSpace = (QWORD)(5LL*1024LL*1024LL);
  return 0;
}
#endif // HAVE_SYS_STATFS_H

// =======================================================
int CImageDisk::mountImageLocation(char * szMountDevice, char * szMountPoint,
   char * szMountFS) // [Main-Thread]
{
  int nRes=0;
  
  errno = 0;
#ifdef OS_LINUX
  nRes = mount(szMountDevice, szMountPoint, szMountFS, 0, NULL);
#endif
#ifdef OS_FBSD
  nRes = mount(szMountFS, szMountPoint, 0, NULL);
#endif

  if ((nRes) && (errno != EBUSY))  // EBUSY==already mounted
    return -1;
  else
    return 0;
}  

// =======================================================
int CImageDisk::umountImageLocation(char *szMountPoint) // [Main-Thread]
{
  sync(); 
  return umount(szMountPoint);
}
